% This function estimates the marginal likelihood of the hybrid TVP-VARs
% in Chan and Eisenstat (2018).
%
% See:
% Chan, J.C.C. and Eisenstat, E. (2018). Comparing Hybrid Time-Varying 
% Parameter VARs, Economics Letters, 171, 1-5.
%

function [ml, mlstd] = ml_hybrid_tvpsv(shortY,store_Sigbeta,store_Sigalp,store_Sigh,...
    store_beta0,store_alp0,store_h0,prior,X2,is_tv,M)
disp('Computing marginal likelihood of a hybrid TVP-SV.... '); 
M = 20*ceil(M/20);
n = size(store_Sigh,2);
k_beta = size(store_Sigbeta,2);
k_alp = size(store_Sigalp,2);
p = (k_beta/n-1)/n;
idx_b_tv = (kron(is_tv,ones(n*p+1,1))==1); % index for time-varying betas
idx_a_tv = false(k_alp,1); % construct index for time-varying alphas
count = 0;
for j=2:n
    if is_tv(j) == 1
        idx_a_tv( count+1:count+j-1) = true;
    end
    count = count + j-1;
end

    % obtain parameters for the optimal importance sampling density
tmp = [zeros(k_beta,1) ones(k_beta,1)];
for i=1:k_beta
    if idx_b_tv(i) == 1
        tmp(i,:) = gamfit(1./store_Sigbeta(:,i));    
    end
end
nubetahat = tmp(:,1); Sbetahat = 1./tmp(:,2);
tmp = [zeros(k_alp,1) ones(k_alp,1)];
for i=1:k_alp
    if idx_a_tv(i) == 1
        tmp(i,:) = gamfit(1./store_Sigalp(:,i)); 
    end
end
nualphat = tmp(:,1); Salphat = 1./tmp(:,2);
tmp = zeros(n,2);
for i=1:n
    tmp(i,:) = gamfit(1./store_Sigh(:,i));    
end
nuhhat = tmp(:,1); Shhat = 1./tmp(:,2);  
beta0hat = mean(store_beta0)';
beta0std = chol(cov(store_beta0),'lower');
beta0pre = cov(store_beta0)\speye(k_beta);
alp0hat = mean(store_alp0)';
alp0std = chol(cov(store_alp0),'lower');
alp0pre = cov(store_alp0)\speye(k_alp);
h0hat = mean(store_h0)';
h0std = chol(cov(store_h0),'lower');
h0pre = cov(store_h0)\speye(n);

    % obtain IS draws
big_Sigbeta = zeros(M,k_beta);
for i=1:k_beta
    if idx_b_tv(i) == 1
        big_Sigbeta(:,i) = 1./gamrnd(nubetahat(i),1./Sbetahat(i),M,1);
    end
end
big_Sigalp = zeros(M,k_alp);
for i=1:k_alp
    if idx_a_tv(i) == 1
        big_Sigalp(:,i) = 1./gamrnd(nualphat(i),1./Salphat(i),M,1);
    end
end
big_Sigh = zeros(M,n);
for i=1:n
    big_Sigh(:,i) = 1./gamrnd(nuhhat(i),1./Shhat(i),M,1);
end
big_beta0 = repmat(beta0hat',M,1) + (beta0std*randn(k_beta,M))';
big_alp0 = repmat(alp0hat',M,1) + (alp0std*randn(k_alp,M))';
big_h0 = repmat(h0hat',M,1) + (h0std*randn(n,M))';

store_w = zeros(M,1);
cIS = nubetahat(idx_b_tv)'*log(Sbetahat(idx_b_tv)) - sum(gammaln(nubetahat(idx_b_tv))) ...
    + nualphat(idx_a_tv)'*log(Salphat(idx_a_tv)) - sum(gammaln(nualphat(idx_a_tv))) ...
    + nuhhat'*log(Shhat) - sum(gammaln(nuhhat)) ...
    -.5*(k_beta+k_alp+n)*log(2*pi) - sum(log(diag(beta0std))) - sum(log(diag(alp0std))) - sum(log(diag(h0std)));
gIS = @(sb,sa,sh,b0,a0,c0) cIS -(nubetahat(idx_b_tv)+1)'*log(sb(idx_b_tv)) - sum(Sbetahat(idx_b_tv)./sb(idx_b_tv)) ... 
    -(nualphat(idx_a_tv)+1)'*log(sa(idx_a_tv)) - sum(Salphat(idx_a_tv)./sa(idx_a_tv)) ...
    -(nuhhat+1)'*log(sh) - sum(Shhat./sh) ...
    -.5*(b0-beta0hat)'*beta0pre*(b0-beta0hat) ...
    -.5*(a0-alp0hat)'*alp0pre*(a0-alp0hat) ...
    -.5*(c0-h0hat)'*h0pre*(c0-h0hat);

for loop = 1:M
    Sigbeta = big_Sigbeta(loop,:)';
    Sigalp = big_Sigalp(loop,:)';
    Sigh = big_Sigh(loop,:)';
    beta0 = big_beta0(loop,:)';
    alp0 = big_alp0(loop,:)';
    h0 = big_h0(loop,:)';
    
    count = 0; llike = 0;
    for ii=1:n        
        if is_tv(ii) == 1
            bigXi = SURform([X2 -shortY(:,1:ii-1)]);
            Sigthetai = [Sigbeta((ii-1)*k_beta/n+1:ii*k_beta/n); Sigalp(count+1:count+ii-1)];        
            thetai0 = [beta0((ii-1)*k_beta/n+1:ii*k_beta/n); alp0(count+1:count+ii-1)];
            llikei = intlike_tvpsv(shortY(:,ii),Sigthetai,Sigh(ii),bigXi,h0(ii),thetai0);
        else        
            bigXi = [X2 -shortY(:,1:ii-1)];
            thetai = [beta0((ii-1)*k_beta/n+1:ii*k_beta/n); alp0(count+1:count+ii-1)];
            llikei = intlike_varsv(shortY(:,ii),thetai,Sigh(ii),bigXi,h0(ii));
        end
        llike = llike + llikei;
        count = count + ii-1;
    end
    store_w(loop) = llike + prior(Sigbeta,Sigalp,Sigh,beta0,alp0,h0) ...
        - gIS(Sigbeta,Sigalp,Sigh,beta0,alp0,h0);
    
    if (mod(loop, 2000) == 0)
        disp([num2str(loop) ' loops... ']);
    end 
end
shortw = reshape(store_w,M/20,20);
maxw = max(shortw);

bigml = log(mean(exp(shortw-repmat(maxw,M/20,1)),1)) + maxw;
ml = mean(bigml);
mlstd = std(bigml)/sqrt(20);

end